########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import yaml
from mock import Mock, call, patch

from . import get_gen
from .. import utils


def test_process_relevant_tokens():

    class Foo(object):
        def __init__(self, keyword):
            self.prev = Mock()
            self.prev.node = Mock(value=keyword)

    @utils.process_relevant_tokens(Foo, 'foo')
    def foo(**kwargs):
        yield kwargs

    tok = Foo('foo')
    assert next(foo(token=tok)) == {'token': tok}

    try:
        tok = Foo('bar')
        next(foo(token=tok))
    except StopIteration:
        pass

    @utils.process_relevant_tokens(Foo, ['baz'])
    def baz(**kwargs):
        yield kwargs

    tok = Foo('baz')
    assert next(baz(token=tok)) == {'token': tok}

    try:
        tok = Foo('bar')
        next(baz(token=tok))
    except StopIteration:
        pass


def test_recurse_mapping():
    gen = get_gen()
    inputs = []
    previous = None
    while True:
        try:
            yaml_mapped = next(gen)
        except StopIteration:
            break
        current = utils.recurse_mapping(yaml_mapped)
        if previous == 'inputs' and isinstance(current, dict):
            inputs.extend(current.keys())
        elif previous == 'get_input' and isinstance(current, str):
            assert current in inputs
        previous = current


def test_setup_node_template():
    gen = get_gen()
    node_templates = []
    while True:
        try:
            yaml_mapped = next(gen)
        except StopIteration:
            break
        if not isinstance(yaml_mapped, yaml.nodes.MappingNode):
            continue
        if not isinstance(yaml_mapped.value, (list, tuple)):
            continue
        if not isinstance(yaml_mapped.value[0], (list, tuple)):
            continue
        if not isinstance(yaml_mapped.value[0][0], yaml.nodes.ScalarNode):
            continue
        if not yaml_mapped.value[0][0].value == 'foo_template':
            continue
        try:
            node_template = utils.setup_node_template(yaml_mapped.value[0])
        except TypeError:
            continue
        if node_template:
            node_templates.append(node_template)
        node_template = None
    assert node_templates[0].name == 'foo_template'
    assert node_templates[0].node_type == 'foo_type'


def test_update_model():
    elem_mock = Mock()
    elem_mock.line_no = 100
    elem_mock.curr = yaml.tokens.ScalarToken(
        'node_templates', 'node_templates', 100, 200)
    elem_mock.nextnext = yaml.tokens.BlockMappingStartToken(201, 300)
    with patch('cfy_lint.yamllint_ext.utils.context') as context:
        utils.update_model(elem_mock)
    assert call.__setitem__(
        'current_top_level', 'node_templates') in context.mock_calls


def test_assign_nested_node_template_level():
    elem_mock = Mock()
    elem_mock.line_no = 100
    elem_mock.curr = yaml.tokens.ScalarToken(
        'properties', 'properties', 100, 200)
    elem_mock.nextnext = yaml.tokens.BlockEntryToken(201, 300)
    assert utils.assign_nested_node_template_level(elem_mock) == 'properties'


def test_assign_current_top_level():
    elem_mock = Mock()
    elem_mock.line_no = 100
    elem_mock.curr = yaml.tokens.ScalarToken(
        'inputs', 'inputs', 100, 200)
    elem_mock.nextnext = yaml.tokens.BlockMappingStartToken(201, 300)
    assert utils.assign_current_top_level(elem_mock) == 'inputs'



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from contextlib import contextmanager


@contextmanager
def filelines(filename):
    with open(filename, 'r') as file:
        lines = file.readlines()
    yield lines
    with open(filename, 'w') as file:
        file.writelines(lines)


def is_list(line):
    if line.lstrip().startswith('-'):
        return True


def get_indented_regex(line, found):
    if is_list(line):
        regex = r'^\s{%s}[\-\s{1}A-Za-z]' % found
    else:
        regex = r'^\s{%s}[A-Za-z]' % found
    return re.compile(regex)



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import yaml

from cfy_lint.yamllint_ext import LintProblem
from cfy_lint.yamllint_ext.generators import CfyNode
from cfy_lint.yamllint_ext.constants import UNUSED_IMPORT_CTX
from cfy_lint.yamllint_ext.utils import (process_relevant_tokens,
                                         INTRINSIC_FNS,
                                         context as ctx)
from cfy_lint.yamllint_ext.rules.constants import (
    GCP_TYPES,
    AWS_TYPES,
    AZURE_TYPES,
    TERRAFORM_TYPES,
    AWS_VALID_KEY,
    AZURE_VALID_KEY,
    TFLINT_SUPPORTED_CONFIGS,
    TERRATAG_SUPPORTED_FLAGS,
    deprecated_node_types,
    REQUIRED_RELATIONSHIPS,
    firewall_rule_gcp,
    security_group_validation_aws,
    security_group_validation_azure,
    security_group_validation_openstack,
)

VALUES = []

ID = 'node_templates'
TYPE = 'token'
CONF = {
    'allowed-values': list(VALUES),
    'check-keys': bool,
    'check-node-types': bool
}
DEFAULT = {
    'allowed-values': ['true', 'false'],
    'check-keys': True,
    'check-node-types': True
}


@process_relevant_tokens(CfyNode, 'node_templates')
def check(token=None, context=None, node_types=None, **_):
    for node_template in token.node.value:
        if not len(node_template) == 2:
            continue
        parsed_node_template = parse_node_template(
            node_template[1], context.get(node_template[0].value))
        remove_node_type_from_context(parsed_node_template.node_type)
        yield from check_deprecated_node_type(
            parsed_node_template,
            parsed_node_template.line or token.line)
        yield from check_intrinsic_functions(
            parsed_node_template.dict,
            parsed_node_template.line or token.line)
        yield from check_client_config(
            parsed_node_template,
            parsed_node_template.line or token.line)
        yield from check_dependent_types(
            parsed_node_template,
            parsed_node_template.line or token.line)
        yield from check_security_group(
              parsed_node_template,
              parsed_node_template.line or token.line)
        yield from check_node_type_imported(
            node_types,
            parsed_node_template,
            parsed_node_template.line or token.line)
        yield from check_terraform(
            parsed_node_template,
            parsed_node_template.line or token.line)
        yield from check_external_resource(
            parsed_node_template,
            parsed_node_template.line or token.line)


def parse_node_template(node_template_mapping, node_template_model):
    node_template_model.set_values(
        recurse_node_template(node_template_mapping))
    node_template_model.line = node_template_mapping.start_mark.line + 1
    return node_template_model


def check_node_type_imported(node_types, model, line):
    node_types = node_types or {}
    if model.node_type not in node_types:
        yield LintProblem(
            line,
            None,
            "unimported node type: {}".format(model.node_type))


def check_deprecated_node_type(model, line):
    if model.node_type in deprecated_node_types:
        yield LintProblem(
            line,
            None,
            "deprecated node type. "
            "Replace usage of {} with {}.".format(
                model.node_type,
                deprecated_node_types[model.node_type]))


def check_intrinsic_functions(data, line):
    if isinstance(data, dict):
        for key, value in data.items():
            if key in INTRINSIC_FNS:
                yield from validate_instrinsic_function(key, value, line)
            else:
                yield from check_intrinsic_functions(value, line)
    elif isinstance(data, list):
        for item in data:
            yield from check_intrinsic_functions(item, line)


def validate_instrinsic_function(key, value, line):
    if key == 'get_input':
        if isinstance(value, list):
            if value[0] not in ctx.get('inputs', {}):
                yield LintProblem(
                    line,
                    None,
                    "get_input references undefined input: {}".format(value[0])
                )
        elif value not in ctx.get('inputs', {}):
            yield LintProblem(
                line,
                None,
                "get_input references undefined input: {}".format(value)
            )
    elif key in ['get_attribute', 'get_property']:
        if value[0] not in ctx.get('node_templates', {}) and \
                value[0] not in ctx.get('imported_node_templates', {}) \
                and value[0] not in ['SELF', 'TARGET', 'SOURCE']:
            yield LintProblem(
                line,
                None,
                "{} references undefined target {}".format(key, value[0])
            )


def recurse_node_template(mapping):
    if isinstance(mapping, yaml.nodes.ScalarNode):
        return mapping.value
    if isinstance(mapping, yaml.nodes.MappingNode):
        mapping_list = []
        for item in mapping.value:
            mapping_list.append(recurse_node_template(item))
        mapping_dict = {}
        for item in mapping_list:
            try:
                mapping_dict[item[0]] = item[1]
            except KeyError:
                mapping_dict.update(item)
        return mapping_dict
    elif isinstance(mapping, tuple):
        if len(mapping) == 2 and isinstance(mapping[0], yaml.nodes.ScalarNode):
            return {
                mapping[0].value: recurse_node_template(mapping[1])
            }
        else:
            new_list = []
            for item in mapping:
                new_list.append(recurse_node_template(item))
            return new_list
    elif isinstance(mapping, yaml.nodes.SequenceNode):
        new_list = []
        for item in mapping.value:
            new_list.append(recurse_node_template(item))
        return new_list


def check_client_config(model, line):
    if model.node_type in GCP_TYPES:
        yield from check_gcp_config(model, line)
    if model.node_type in AZURE_TYPES:
        yield from check_azure_config(model, line)
    if model.node_type in AWS_TYPES:
        yield from check_aws_config(model, line)


def check_gcp_config(model, line):
    if 'gcp_config' in model.properties:
        yield LintProblem(
            line,
            None,
            'The node template "{}" has deprecated property "gcp_config". '
            'please use "client_config".'.format(model.name)
        )
    elif 'client_config' not in model.properties:
        yield LintProblem(
            line,
            None,
            'The node template "{}" '
            'does not provide required property "client_config".'.format(
                model.name)
        )
    elif not all(x in ['auth', 'zone']
                 for x in model.properties['client_config']):
        yield LintProblem(
            line,
            None,
            'The node template "{}" '
            'does not provide required client config values '
            '["auth", "zone"].'.format(
                model.name)
        )


def check_azure_config(model, line):
    if 'client_config' not in model.properties or \
            'azure_config' in model.properties:
        yield LintProblem(
            line,
            None,
            'The node template "{}" '
            'does not provide required property "client_config".'.format(
                model.name))
    client_config = model.properties.get('client_config', {})
    if not all(x in AZURE_VALID_KEY for x in client_config.keys()):
        if 'get_input' not in model.properties['client_config'] and \
                'get_secret' not in model.properties['client_config']:
            yield LintProblem(line,
                              None,
                              'Invalid parameters provided for client config. '
                              'Valid parameters are {}'.format(
                                  AZURE_VALID_KEY))


def check_aws_config(model, line):
    if 'client_config' not in model.properties:
        yield LintProblem(
            line,
            None,
            'The node template "{}" '
            'does not provide required property "client_config".'.format(
                model.name))
    client_config = model.properties.get('client_config')
    if not all(x in AWS_VALID_KEY for x in client_config.keys()):
        if 'get_input' not in model.properties['client_config'] and \
                'get_secret' not in model.properties['client_config']:
            yield LintProblem(line,
                              None,
                              'Invalid parameters provided for client config. '
                              'Valid parameters are {}'.format(AWS_VALID_KEY))


def check_dependent_types(model, line):
    required_relationship_types = REQUIRED_RELATIONSHIPS.get(
        model.node_type, {})
    model.required_relationships = required_relationship_types
    if model.required_relationships_not_met(ctx['node_templates'],
                                            ctx.get('imported_node_templates',
                                                    {})):
        yield LintProblem(
            line,
            None,
            model.required_relationships_message
        )


def check_security_group(model, line):
    if model.node_type in security_group_validation_aws:
        yield from check_security_group_validation_aws(model, line)
    if model.node_type in security_group_validation_azure:
        yield from check_security_group_validation_azure(model, line)
    if model.node_type in security_group_validation_openstack:
        yield from check_security_group_validation_openstack(model, line)
    if model.node_type in firewall_rule_gcp:
        yield from check_firewall_rule_gcp(model, line)


def check_security_group_validation_aws(model, line):
    resource_config = model.properties.get('resource_config', {})
    ip_permissions = resource_config.get('IpPermissions', {})
    for item in ip_permissions:
        from_port = item.get('FromPort', {})
        to_port = item.get('ToPort', {})
        if from_port == '-1' or to_port == '-1':
            yield LintProblem(
                line,
                None,
                "Security group rule too open. {}".format(item))
        if int(to_port) - int(from_port) < 0:
            yield LintProblem(
                line,
                None,
                "Security group The port is invalid. {}".format(item))


def check_security_group_validation_azure(model, line):
    resource_config = model.properties.get('resource_config', {})
    security_rules = resource_config.get('securityRules', {})
    for item in security_rules:
        destination_port_range = item['properties'].get(
            'destinationPortRange', {})
        if destination_port_range == '*':
            yield LintProblem(
                line,
                None,
                "Security group rule too open. {}".format(item))


def check_security_group_validation_openstack(model, line):
    security_group_rules = model.properties.get('security_group_rules', {})
    for item in security_group_rules:
        protocol = item.get('protocol', {})
        port_range_min = item.get('port_range_min', {})
        port_range_max = item.get('port_range_max', {})
        if port_range_max == 'null' or port_range_min == 'null':
            if protocol != 'icmp':
                yield LintProblem(
                    line,
                    None,
                    "Security group rule Invalid. {}".format(item))
        elif port_range_max == '65535' or port_range_min == '1':
            yield LintProblem(
                line,
                None,
                "Security group rule too open. {}".format(item))
        elif int(port_range_max) - int(port_range_min) < 0:
            yield LintProblem(
                line,
                None,
                "Security group The port range is invalid. {}".format(item))


def check_firewall_rule_gcp(model, line):
    allowed = model.properties.get('allowed', {})
    for item in allowed['tcp']:
        if '-' in str(item):  # 12345-12349
            ports = re.split('-', item)
            if int(ports[0]) > int(ports[1]):
                yield LintProblem(
                    line,
                    None,
                    "Security group The port range is invalid.{}".format(item))


def check_terraform(model, line):
    if model.node_type in TERRAFORM_TYPES:
        tflint_config = model.properties.get('tflint_config', {})
        tfsec_config = model.properties.get('tfsec_config', {})
        terratag_config = model.properties.get('terratag_config', {})
        if tflint_config:
            yield from check_tflint(model, line)
        if tfsec_config:
            yield from check_tfsec(model, line)
        if terratag_config:
            yield from check_terratag(model, line)


def check_tfsec(model, line):
    tfsec_config = model.properties.get('tfsec_config', {})
    enable = tfsec_config.get('enable', {})
    if enable and enable == 'false':
        yield LintProblem(
            line,
            None,
            'tfsec_config will have no effect if "enable: false".')
    config = tfsec_config.get('config', {})
    include = config.get('include', None)
    exclude = config.get('exclude', None)
    if exclude and not isinstance(exclude, list) or \
            include and not isinstance(include, list):
        yield LintProblem(
            line,
            None,
            'tfsec_config.config '
            'parameters "include" and "exclude" should be a list')
    flags_override = tfsec_config.get('flags_override', {})
    for flag in flags_override:
        if 'color' == flag:
            yield LintProblem(
                line,
                None,
                'Color flag cannot be used in flags_override')


def check_tflint(model, line):
    tflint_config = model.properties.get('tflint_config', {})
    enable = tflint_config.get('enable', {})
    if enable and enable == 'false':
        yield LintProblem(
            line,
            None,
            'tflint_config will have no effect if "enable: false".')
    config = tflint_config.get('config', {})
    for item in config:
        type_name = item.get('type_name', {})
        if type_name not in TFLINT_SUPPORTED_CONFIGS:
            yield LintProblem(
                line,
                None,
                'unsupported key {} in tflint_config.'
                .format(TFLINT_SUPPORTED_CONFIGS))
        option_name = item.get('option_name', {})
        if type_name == 'plugin' and not option_name:
            yield LintProblem(
                line,
                None,
                'tflint_config "type_name" key must also provide '
                '"option_name", which is the plugin name.')
        elif type_name == 'config':
            option_value = item.get('option_value', {})
            if not option_value:
                yield LintProblem(
                    line,
                    None,
                    'To use tflint with type_name: config, it is necessary to '
                    'write option_value ')
    flags_override = tflint_config.get('flags_override', {})
    for flag in flags_override:
        if 'color' == flag:
            yield LintProblem(
                line,
                None,
                'color flag is not supported in flags_override')


def check_terratag(model, line):
    terratag_config = model.properties.get('terratag_config', {})
    enable = terratag_config.get('enable', {})
    if enable and enable == 'false':
        yield LintProblem(
            line,
            None,
            'terratag_config will have no effect if "enable: false".')
    tags = terratag_config.get('tags', {})
    if not tags or not isinstance(tags, dict):
        yield LintProblem(
            line,
            None,
            'tags should be a dict')
    flags_override = terratag_config.get('flags_override', {})
    if not isinstance(flags_override, list):
        yield LintProblem(
            line,
            None,
            'flags_override should be a list')
    for flag in flags_override:
        if not isinstance(flag, dict):
            yield LintProblem(
                line,
                None,
                'The flags inside flags_override should be a dict')
        key = flag.keys()
        for key in key:
            if '-' in key:
                yield LintProblem(
                    line,
                    None,
                    'The flags should be without a "-" sign, {}'.format(key))
            if key not in TERRATAG_SUPPORTED_FLAGS:
                yield LintProblem(
                    line,
                    None,
                    'unsupported flag, {}'.format(TERRATAG_SUPPORTED_FLAGS))


def check_external_resource(model, line):
    if model.is_external and 'resource_config' in model.properties:
        yield LintProblem(
            line,
            None,
            'resource_config is not required, '
            'when use_external_resource is true.')


def remove_node_type_from_context(node_type):
    if UNUSED_IMPORT_CTX in ctx:
        for import_item in list(ctx[UNUSED_IMPORT_CTX].keys()):
            if node_type in ctx[UNUSED_IMPORT_CTX][import_item]:
                del ctx[UNUSED_IMPORT_CTX][import_item]



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from cfy_lint.yamllint_ext.autofix.utils import (filelines, get_indented_regex)


def fix_indentation(problem):
    if problem.rule == 'indentation':
        with filelines(problem.file) as lines:
            expected, found = get_space_diff(problem.desc)
            indented_regex = get_indented_regex(
                lines[problem.line - 1], len(found))
            idx = problem.line - 1
            while True:
                if idx == len(lines) or not indented_regex.match(lines[idx]):
                    break
                lines[idx] = replace_spaces(expected, found, lines[idx])
                idx += 1
                continue
        problem.fixed = True


def replace_spaces(expected, found, line):
    sans_newline = line.rstrip()
    new_line = sans_newline.replace(found, expected)
    if sans_newline != line:
        new_line += '\n'
    return new_line


def get_space_diff(message):
    """Get the indentation that exists, and what we want to replace it with.
    :param message: The string message that we have in problem.desc
    :return: (found, desired)
    """
    found_expect = re.findall(r'\d+', message)
    if isinstance(found_expect, list) and len(found_expect) == 2:
        return int(found_expect[0]) * ' ', int(found_expect[1]) * ' '
    return 0, 0



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import os
import sys
import json
from re import sub
from logging import (Formatter, StreamHandler)

from cfy_lint import cli
from cfy_lint.yamllint_ext import (run, rules)
from cfy_lint.logger import logger, stream_handler
from cfy_lint.yamllint_ext.config import YamlLintConfigExt


@cli.command()
@cli.options.blueprint_path
@cli.options.config
@cli.options.verbose
@cli.options.format
@cli.options.skip_suggestions
@cli.options.autofix
def lint(blueprint_path,
         config,
         verbose,
         format,
         skip_suggestions=None,
         autofix=False):

    yaml_config = YamlLintConfigExt(content=config, yamllint_rules=rules)
    skip_suggestions = skip_suggestions or ()
    try:
        report = create_report_for_file(
            blueprint_path,
            yaml_config,
            skip_suggestions=skip_suggestions,
            autofix=autofix)
    except Exception as e:
        if verbose:
            raise e
        else:
            exception_str = str(e)
        logger.error(exception_str)
        sys.exit(1)

    if format == 'json':
        logger.removeHandler(stream_handler)
        new_logging_handler = StreamHandler()
        new_logging_formatter = Formatter(fmt='%(message)s')
        new_logging_handler.setFormatter(new_logging_formatter)
        logger.addHandler(new_logging_handler)

    cnt = 0
    for item in report:
        message = formatted_message(item, format)
        if cnt == 0:
            logger.info('The following linting errors were found: ')
            cnt += 1
        if item.level == 'warning':
            logger.warning(message)
        elif item.level == 'error':
            logger.error(message)
        else:
            logger.info(message)


def create_report_for_file(file_path,
                           conf,
                           create_report_for_file=False,
                           skip_suggestions=None,
                           autofix=False):
    if not os.path.exists(file_path):
        raise RuntimeError('File path does not exist: {}.'.format(file_path))
    logger.info('Linting blueprint: {}'.format(file_path))
    with io.open(file_path, newline='') as f:
        return run(f, conf, create_report_for_file, skip_suggestions, autofix)


def formatted_message(item, format=None):
    if format == 'json':
        rule, item_message = item.message.split(':', 1)
        return json.dumps({
            "level": item.level,
            "line": item.line,
            "rule": sub(r"[()]", "", rule),
            "message": item_message,
        })
    return '{0: <4}: {1:>4}'.format(item.line, item.message)



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import yaml


class CloudifyDSLObject(object):

    def __init__(self, tokens=None):
        self._tokens = tokens or []

    @property
    def tokens(self):
        return self._tokens

    @tokens.setter
    def tokens(self, value):
        self._tokens = value

    def seek(self, token_class, index):
        while True:
            try:
                if isinstance(self.tokens[index], token_class):
                    return index
            except IndexError:
                return
            index += 1


class NodeTemplate(object):
    def __init__(self, name):
        self.name = name
        self._type = None
        self._properties = None
        self._interfaces = None
        self._relationships = []
        self._line = None
        self._required_relationships = {}
        self._relationships_mapping = {}
        self._unsatisfied_relationships = {}

    @property
    def node_type(self):
        return self._type

    @node_type.setter
    def node_type(self, value):
        self._type = value

    @property
    def properties(self):
        return self._properties

    @properties.setter
    def properties(self, value):
        self._properties = value

    @property
    def interfaces(self):
        return self._interfaces

    @interfaces.setter
    def interfaces(self, value):
        self._interfaces = value

    @property
    def relationships(self):
        return self._relationships

    @relationships.setter
    def relationships(self, value):
        self._relationships = value

    @property
    def is_external(self):
        if self.properties:
            return self.properties.get(
                'use_external_resource', False)
        return False

    def set_values(self, values):
        if 'type' in values:
            self.node_type = values['type']
        if 'properties' in values:
            self.properties = values['properties']
        if 'interfaces' in values:
            self.interfaces = values['interfaces']
        if 'relationships' in values:
            self.relationships = values['relationships']

    @property
    def is_third_party(self):
        if '.aws.' in self.node_type:
            return True
        elif '.azure.' in self.node_type:
            return True
        elif '.gcp.' in self.node_type:
            return True
        return False

    @property
    def dict(self):
        return {
            self.name: {
                'type': self.node_type,
                'properties': self.properties,
                'interfaces': self.interfaces,
                'relationships': self.relationships
            }
        }

    @property
    def line(self):
        return self._line

    @line.setter
    def line(self, value):
        self._line = value

    @property
    def required_relationships(self):
        return self._required_relationships or {}

    @required_relationships.setter
    def required_relationships(self, value):
        self._required_relationships = value

    def required_relationships_not_met(self,
                                       node_templates=None,
                                       imported=None):
        node_templates = node_templates or {}
        if self.is_external or self.is_third_party:
            return self._unsatisfied_relationships.items()
        if imported:
            for k, v in imported.items():
                node_templates[k] = NodeTemplate(v)
        if not self._relationships_mapping:
            for relationship in self.relationships:
                if isinstance(relationship, dict):
                    rel_type = relationship.get('type')
                    rel_target_name = relationship.get('target')
                    rel_target = node_templates.get(
                        rel_target_name, UnknownNodeType(rel_target_name))
                    self._relationships_mapping[rel_target.node_type] = \
                        rel_type
        for k, v in self.required_relationships.items():
            if k not in self._relationships_mapping or v != \
                    self._relationships_mapping[k]:
                self._unsatisfied_relationships[k] = v
        return self._unsatisfied_relationships.items()

    @property
    def required_relationships_message(self):
        messages = []
        for k, v in self.required_relationships_not_met():
            messages.append(
                'relationship type {} to a node type {}'.format(v, k))
        return 'The node template {} has unsatisfied required relationships' \
               ', which have not been provided: {}.'.format(
                   self.name, ', '.join(messages))


class RelationshipsList(CloudifyDSLObject):

    kind = 'relationships'

    def __init__(self, current, tokens):
        self.current = current
        super().__init__(tokens)
        self._relationship_items = []

    def setup(self):
        index = 0
        limit = len(self.tokens)
        while not index >= limit:
            if isinstance(self.tokens[index],
                          yaml.tokens.BlockMappingStartToken):
                block_end = self.seek(yaml.tokens.BlockEndToken, index + 1)
                if block_end:
                    self._relationship_items.append(
                        [RelationshipItem(self.tokens[index:block_end])])
                index = block_end
            else:
                index += 1

    @property
    def relationship_items(self):
        return self._relationship_items


class RelationshipItem(CloudifyDSLObject):
    kind = ('type', 'target')

    def __init__(self, tokens):
        self._type = None
        self._target = None
        super().__init__(tokens)
        self.setup()

    def setup(self):
        index = 0
        limit = len(self.tokens)
        while not self._type and not self._target:
            if index >= limit:
                break
            if isinstance(self.tokens[index], yaml.tokens.ScalarToken):
                if isinstance(self.tokens[index + 2], yaml.tokens.ScalarToken):
                    if self.tokens[index].value == 'type':
                        self._type = self.tokens[index + 2]
                    elif self.tokens[index].value == 'target':
                        self._target = self.tokens[index + 2]
            index += 1

    def validate(self):
        return self._type and self._target


class UnknownNodeType(NodeTemplate):
    def __init__(self, name):
        super().__init__(name)
        self.node_type = 'Unknown Node Type'



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cfy_lint.yamllint_ext.autofix.truthy import fix_truthy
from cfy_lint.yamllint_ext.autofix.colons import fix_colons
from cfy_lint.yamllint_ext.autofix.indentation import fix_indentation
from cfy_lint.yamllint_ext.autofix.brackets import fix_spaces_in_brackets
from cfy_lint.yamllint_ext.autofix.trailing_spaces import fix_trailing_spaces
from cfy_lint.yamllint_ext.autofix.deprecated_node_types import \
    fix_deprecated_node_types
from cfy_lint.yamllint_ext.autofix.deprecated_relationships import \
    fix_deprecated_relationships


def fix_problem(problem):
    if problem.file or problem.line:
        fix_truthy(problem)
        fix_indentation(problem)
        fix_trailing_spaces(problem)
        fix_deprecated_node_types(problem)
        fix_deprecated_relationships(problem)
        fix_colons(problem)
        fix_spaces_in_brackets(problem)



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re

from yamllint import parser
from cfy_lint.yamllint_ext.constants import (
    UNUSED_INPUTS, UNUSED_IMPORT_CTX)
from cfy_lint.yamllint_ext.generators import (
    CfyNode,
    CfyToken,
    YamlParserLintProblem,
    token_or_comment_or_line_generator,
)
from cfy_lint.yamllint_ext.overrides import (
    LintProblem,
    spaces_after,
    spaces_before,
    get_syntax_error
)
from cfy_lint.yamllint_ext.utils import (
    context,
    setup_types,
    update_model,
    setup_node_templates,
)
from cfy_lint.yamllint_ext.autofix import fix_problem
from cfy_lint.yamllint_ext.rules.inputs import ID as input_rule
from cfy_lint.yamllint_ext.rules.imports import ID as import_rule
from cfy_lint.yamllint_ext.autofix.add_label import fix_add_label
from cfy_lint.yamllint_ext.autofix.empty_lines import fix_empty_lines

PROBLEM_LEVELS = {
    0: None,
    1: 'warning',
    2: 'error',
    None: 0,
    'warning': 1,
    'error': 2,
}


def get_cosmetic_problems(buffer,
                          conf,
                          filepath,
                          base_path=None,
                          skip_suggestions=None):
    setup_types(buffer, base_path=base_path)

    rules = conf.enabled_rules(filepath)

    # Split token rules from line rules
    token_rules = [r for r in rules if r.TYPE == 'token']
    comment_rules = [r for r in rules if r.TYPE == 'comment']
    line_rules = [r for r in rules if r.TYPE == 'line']

    for rule in token_rules:
        context[rule.ID] = {}

    class DisableDirective:
        def __init__(self):
            self.rules = set()
            self.all_rules = {r.ID for r in rules}

        def process_comment(self, comment):
            try:
                comment = str(comment)
            except UnicodeError:
                return  # this certainly wasn't a yamllint directive comment

            if re.match(r'^# yamllint disable( rule:\S+)*\s*$', comment):
                items = comment[18:].rstrip().split(' ')
                rules = [item[5:] for item in items][1:]
                if len(rules) == 0:
                    self.rules = self.all_rules.copy()
                else:
                    for id in rules:
                        if id in self.all_rules:
                            self.rules.add(id)

            elif re.match(r'^# yamllint enable( rule:\S+)*\s*$', comment):
                items = comment[17:].rstrip().split(' ')
                rules = [item[5:] for item in items][1:]
                if len(rules) == 0:
                    self.rules.clear()
                else:
                    for id in rules:
                        self.rules.discard(id)

        def is_disabled_by_directive(self, problem):
            return problem.rule in self.rules

    class DisableLineDirective(DisableDirective):
        def process_comment(self, comment):
            try:
                comment = str(comment)
            except UnicodeError:
                return  # this certainly wasn't a yamllint directive comment

            if re.match(r'^# yamllint disable-line( rule:\S+)*\s*$', comment):
                items = comment[23:].rstrip().split(' ')
                rules = [item[5:] for item in items][1:]
                if len(rules) == 0:
                    self.rules = self.all_rules.copy()
                else:
                    for id in rules:
                        if id in self.all_rules:
                            self.rules.add(id)

    # Use a cache to store problems and flush it only when a end of line is
    # found. This allows the use of yamllint directive to disable some rules on
    # some lines.
    cache = []
    disabled = DisableDirective()
    disabled_for_line = DisableLineDirective()
    disabled_for_next_line = DisableLineDirective()

    for elem in token_or_comment_or_line_generator(buffer):
        if isinstance(elem, YamlParserLintProblem):
            yield LintProblem(elem.line, elem. column, elem.desc)

        if isinstance(elem, CfyNode):
            setup_node_templates(elem)
            for rule in token_rules:
                rule_conf = conf.rules[rule.ID]
                elem.blueprint_path = base_path
                try:
                    problems = rule.check(
                        conf=rule_conf,
                        token=elem,
                        context=context[rule.ID],
                        node_types=context.get('imported_node_types', {}),
                        skip_suggestions=skip_suggestions)
                except TypeError:
                    continue
                else:
                    for problem in problems:
                        problem.rule = rule.ID
                        problem.level = rule_conf['level']
                        cache.append(problem)

        elif isinstance(elem, CfyToken):
            update_model(elem)
            for rule in token_rules:
                if hasattr(rule, 'LintProblem'):
                    rule.LintProblem = LintProblem
                if hasattr(rule, 'spaces_before'):
                    rule.spaces_before = spaces_before
                if hasattr(rule, 'spaces_after'):
                    rule.spaces_after = spaces_after
                rule_conf = conf.rules[rule.ID]
                problems = rule.check(rule_conf,
                                      elem.curr,
                                      elem.prev,
                                      elem.after,
                                      elem.nextnext,
                                      context[rule.ID])
                for problem in problems:
                    problem.rule = rule.ID
                    problem.level = rule_conf['level']
                    cache.append(problem)

        elif isinstance(elem, parser.Comment):
            for rule in comment_rules:
                if hasattr(rule, 'LintProblem'):
                    rule.LintProblem = LintProblem
                if hasattr(rule, 'spaces_before'):
                    rule.spaces_before = spaces_before
                if hasattr(rule, 'spaces_after'):
                    rule.spaces_after = spaces_after
                rule_conf = conf.rules[rule.ID]
                for problem in rule.check(rule_conf, elem):
                    problem.rule = rule.ID
                    problem.level = rule_conf['level']
                    cache.append(problem)

            disabled.process_comment(elem)
            if elem.is_inline():
                disabled_for_line.process_comment(elem)
            else:
                disabled_for_next_line.process_comment(elem)

        elif isinstance(elem, parser.Line):
            for rule in line_rules:
                if hasattr(rule, 'LintProblem'):
                    rule.LintProblem = LintProblem
                if hasattr(rule, 'spaces_before'):
                    rule.spaces_before = spaces_before
                if hasattr(rule, 'spaces_after'):
                    rule.spaces_after = spaces_after
                rule_conf = conf.rules[rule.ID]
                for problem in rule.check(rule_conf, elem):
                    problem.rule = rule.ID
                    problem.level = rule_conf['level']
                    cache.append(problem)

    # if this runs with the other loop, we will not have a chance
    # to check usage of plugins in the blueprint, before
    # it is reported.
    for import_item, problem in context.get(
            'post_processing_problems', {}).items():
        if import_item in context[UNUSED_IMPORT_CTX]:
            problem.rule = import_rule
            problem.level = 'error'
            cache.append(problem)

    for _, problem in context.get(UNUSED_INPUTS, {}).items():
        problem.rule = input_rule
        problem.level = 'error'
        cache.append(problem)

    # This is the last token/comment/line of this line, let's flush the
    # problems found (but filter them according to the directives)
    for problem in cache:
        if not (disabled_for_line.is_disabled_by_directive(problem) or
                disabled.is_disabled_by_directive(problem)):
            yield problem

    disabled_for_line = disabled_for_next_line
    disabled_for_next_line = DisableLineDirective()
    cache = []


def _run(buffer,
         conf,
         filepath,
         base_path=None,
         input_file=None,
         skip_suggestions=None,
         autofix=False):

    assert hasattr(buffer, '__getitem__'), \
        '_run() argument must be a buffer, not a stream'

    first_line = next(parser.line_generator(buffer)).content
    if re.match(r'^#\s*yamllint disable-file\s*$', first_line):
        return

    # If the document contains a syntax error, save it and yield it at the
    # right line
    syntax_error = get_syntax_error(buffer)

    problems = list(get_cosmetic_problems(buffer,
                                          conf,
                                          filepath,
                                          base_path,
                                          skip_suggestions))

    sorted_problems = sorted(problems, key=lambda x: x.line)
    add_label = False
    extra_empty_line = False
    for problem in sorted_problems:
        # Insert the syntax error (if any) at the right place...

        if (syntax_error and syntax_error.line <= problem.line and
                syntax_error.column <= problem.column):
            yield syntax_error

            # If there is already a yamllint error at the same place, discard
            # it as it is probably redundant (and maybe it's just a 'warning',
            # in which case the script won't even exit with a failure status).
            if (syntax_error.line == problem.line and
                    syntax_error.column == problem.column):
                syntax_error = None
                continue

            syntax_error = None

        if autofix:
            input_file_path = os.path.abspath(input_file)
            problem.file = input_file_path

            if problem.rule == 'inputs':
                add_label = True
                problem.fixed = True
            if problem.rule == 'empty-lines':
                extra_empty_line = True
                problem.fixed = True
            fix_problem(problem)

        if not problem.fixed:
            yield problem

    if autofix and add_label:
        fix_add_label(sorted_problems)

    # this needs to be separated from the rest of the auto fix functions since
    # it changes the line numbers of the entire file, so we do it once all
    # other tasks are done
    if autofix and extra_empty_line:
        fix_empty_lines(problem)

    if syntax_error:
        yield syntax_error


def run(input, conf, filepath=None, skip_suggestions=None, autofix=False):
    """Lints a YAML source.

    Returns a generator of LintProblem objects.

    :param input: buffer, string or stream to read from
    :param conf: yamllint configuration object
    :param filepath: The config file path.
    :param skip_suggestions: Do not suggest changes in lint message.
    :param autofix: fix changes in place.
    """

    if conf.is_file_ignored(filepath):
        return ()

    base_path = os.path.dirname(os.path.abspath(input.name))

    if isinstance(input, (bytes, str)):
        base_path = os.path.dirname(os.path.abspath(input.name))
        return _run(input,
                    conf,
                    filepath,
                    input_file=input.name,
                    base_path=base_path,
                    skip_suggestions=skip_suggestions,
                    autofix=autofix)
    elif hasattr(input, 'read'):  # Python 2's file or Python 3's io.IOBase
        # We need to have everything in memory to parse correctly
        content = input.read()
        return _run(content,
                    conf,
                    filepath,
                    input_file=input.name,
                    base_path=base_path,
                    skip_suggestions=skip_suggestions,
                    autofix=autofix)
    else:
        raise TypeError('input should be a string or a stream')



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from mock import Mock, patch

from . import get_loader, get_gen_as_list

from .. import rules
from .. import LintProblem
from ..cloudify import models
from ..generators import (
    CfyNode,
    generate_nodes_recursively)
from ..rules.constants import (
    TFLINT_SUPPORTED_CONFIGS,
    TERRATAG_SUPPORTED_FLAGS)


def get_mock_cfy_node(content, top_level_type, curr_node_index=1):
    loaded_yaml = get_loader(content)
    loaded_yaml.check_node()
    curr_node = get_gen_as_list(
        generate_nodes_recursively, loaded_yaml.get_node().value)
    node = Mock()
    node.start_mark = Mock(line=100)
    node.end_mark = Mock(line=200)
    node.value = curr_node[curr_node_index].value
    prev = Mock(node=Mock(value=top_level_type))
    elem = CfyNode(node, prev)
    elem.line = 1
    return elem


def test_capabilities():
    capability_content = """
    capabilities:
      key_content:
        description: Private agent key
        shmalue: { get_attribute: [agent_key, private_key_export] }
    """
    elem = get_mock_cfy_node(capability_content, 'capabilities')
    result = get_gen_as_list(rules.capabilities.check, {'token': elem})
    assert isinstance(result[0], LintProblem)
    assert "capability key_content does not provide a value" in \
           result[0].message

    output_content = """
    outputs:
      key_content:
        description: Private agent key
        shmalue: { get_attribute: [agent_key, private_key_export] }
    """
    elem = get_mock_cfy_node(output_content, 'outputs')
    result = get_gen_as_list(rules.capabilities.check, {'token': elem})
    assert isinstance(result[0], LintProblem)
    assert "output key_content does not provide a value" in \
           result[0].message


def test_dsl_definition():
    dsl_def_content_a = """
    dsl_definitions:
      1: &foo
        foo: bar
    """
    elem = get_mock_cfy_node(dsl_def_content_a, 'dsl_definitions')
    result = get_gen_as_list(rules.dsl_definitions.check, {'token': elem})
    assert isinstance(result[0], LintProblem)
    assert "dsl definition should be a string and " \
           "should not start with a numeric character" in result[0].message

    dsl_def_content_2 = """
    dsl_definitions:
      foo: &foo
        - foo
    """
    elem = get_mock_cfy_node(dsl_def_content_2, 'dsl_definitions')
    result = get_gen_as_list(rules.dsl_definitions.check, {'token': elem})
    assert isinstance(result[0], LintProblem)
    assert "dsl definition foo content must be a dict" in \
           result[0].message


def test_dsl_versions():
    dsl_version_content = """
    tosca_definitions_version: cloudify_dsl_1_5
    """
    elem = get_mock_cfy_node(dsl_version_content, 'tosca_definitions_version')
    result = get_gen_as_list(rules.dsl_version.check, {'token': elem})
    assert isinstance(result[0], LintProblem)
    assert "dsl_version not supported: cloudify_dsl_1_5" in \
           result[0].message


def test_imports():
    output_content = """
    imports:
      - ftp://cloudify.co/spec/cloudify/6.3.0/types.yaml
      - plugin:cloudify-openstack-plugin?version= <=3.0.0
    """
    elem = get_mock_cfy_node(output_content, 'imports')
    result = get_gen_as_list(rules.imports.check, {'token': elem})
    assert isinstance(result[0], LintProblem)
    assert "ftp scheme not accepted" in result[0].message
    assert isinstance(result[1], LintProblem)
    assert "Cloudify Openstack Plugin version ['version= <=3.0.0']" in \
           result[1].message


def test_inputs():
    input_content = """
    inputs:
      taco:
        description: taco
        default: 'taco'
    """
    elem = get_mock_cfy_node(input_content, 'inputs')
    result = get_gen_as_list(rules.inputs.check, {'token': elem,
                                                  'skip_suggestions': ()})
    assert isinstance(result[0], LintProblem)
    assert '"taco" does not specify a type. ' \
           'The correct type could be "string".' in result[0].message
    input_content_2 = """
    node_templates:
      foo:
        type: cloudify.nodes.Foo
        properties:
          bar: { get_input: baz }
    """

    elem = get_mock_cfy_node(input_content_2, 'get_input')
    with patch('cfy_lint.yamllint_ext.rules.inputs.ctx') as ctx:
        ctx['inputs'] = {}
        result = get_gen_as_list(rules.inputs.check, {'token': elem})
        assert isinstance(result[0], LintProblem)
        assert 'undefined input' in result[0].message


def test_node_templates():
    node_templates_content = """
    node_templates:

      foo:
        type: cloudify.azure.nodes.ResourceGroup
        properties:
          azure_config:
            foo: bar
          baz: { get_input: taco }
          quk: { get_attribute: [ quuk, quuz ] }
    """

    elem = get_mock_cfy_node(node_templates_content, 'node_templates')
    context = {
        'foo': models.NodeTemplate('foo'),
    }
    with patch('cfy_lint.yamllint_ext.rules.node_templates.ctx') as ctx:
        ctx['inputs'] = {}
        result = get_gen_as_list(rules.node_templates.check,
                                 {'token': elem, 'context': context})

        assert isinstance(result[0], LintProblem)
        assert 'deprecated node type' in result[0].message
        assert isinstance(result[1], LintProblem)
        assert 'undefined input' in result[1].message
        assert isinstance(result[2], LintProblem)
        assert 'undefined target' in result[2].message
        assert isinstance(result[3], LintProblem)
        assert 'required property' in result[3].message


def test_node_types():
    node_types_content = """
    node_types:
      foo:
        derived_from: cloudify.nodes.Root
    """
    elem = get_mock_cfy_node(node_types_content, 'node_types')
    result = get_gen_as_list(rules.node_types.check, {'token': elem,
                                                      'skip_suggestions': ()})
    assert 'naming convention cloudify.nodes.*' in result[0].message


def test_relationships():
    relationships_content = """
    node_templates:
      foo:
        type: cloudify.nodes.Foo
        relationships:
        - type: cloudify.azure.relationships.contained_in_resource_group
          target: foo
    """
    # elem = get_mock_cfy_node(relationships_content, 'relationships')
    loaded_yaml = get_loader(relationships_content)
    loaded_yaml.check_node()
    curr_node = get_gen_as_list(
        generate_nodes_recursively, loaded_yaml.get_node().value)
    prev = Mock(node=Mock(value='relationships'))
    elem = CfyNode(curr_node[7], prev)
    elem.line = 1
    elem.node_templates = ['foo', 'bar']
    result = get_gen_as_list(rules.relationships.check, {'token': elem})

    assert 'deprecated relationship type' in result[0].message


def test_tflint():
    node_templates_content = """
    node_templates:
      cloud_resources:
        type: cloudify.nodes.terraform.Module
        properties:
          tflint_config:
            config:
            - type_name: configinvalid
            - type_name: config
              option_value_invalid:
                module: 'true'
            - type_name: plugin
              option_name_invalid: aws
              option_value:
                enabled: 'false'
            flags_override:
              - loglevel: info
              - color
            enable: false
    """

    elem = get_mock_cfy_node(node_templates_content, 'node_templates')
    context = {
        'cloud_resources': models.NodeTemplate('cloud_resources'),
    }

    with patch('cfy_lint.yamllint_ext.rules.node_templates.ctx') as ctx:
        ctx['inputs'] = {}
        result = get_gen_as_list(
            rules.node_templates.check,
            {
                'token': elem,
                'context': context,
                'node_types': ['cloudify.nodes.terraform.Module']
            }
        )
        result.pop(0)
        assert isinstance(result[0], LintProblem)
        assert 'tflint_config will have no effect if "enable: false".' \
               in result[0].message
        assert isinstance(result[1], LintProblem)
        assert 'unsupported key {} in tflint_config.'\
            .format(TFLINT_SUPPORTED_CONFIGS) in result[1].message
        assert isinstance(result[2], LintProblem)
        assert 'To use tflint with type_name: config, it is necessary to ' \
               'write option_value' in result[2].message
        assert isinstance(result[3], LintProblem)
        assert 'tflint_config "type_name" key must also provide ' \
               '"option_name", which is the plugin name.' in result[3].message
        assert isinstance(result[4], LintProblem)
        assert 'color flag is not supported in flags_override'\
               in result[4].message


def test_tfsec():
    node_templates_content = """
    node_templates:
      cloud_resources:
        type: cloudify.nodes.terraform.Module
        properties:
          tfsec_config:
            config:
              "exclude" : 'invalid'
            flags_override: [color]
            enable: false
    """

    elem = get_mock_cfy_node(node_templates_content, 'node_templates')
    context = {
        'cloud_resources': models.NodeTemplate('cloud_resources'),
    }

    with patch('cfy_lint.yamllint_ext.rules.node_templates.ctx') as ctx:
        ctx['inputs'] = {}
        result = get_gen_as_list(
            rules.node_templates.check,
            {
                'token': elem,
                'context': context,
                'node_types': ['cloudify.nodes.terraform.Module']
            }
        )
        result.pop(0)
        assert isinstance(result[0], LintProblem)
        assert 'tfsec_config will have no effect if "enable: false".' \
               in result[0].message
        assert isinstance(result[1], LintProblem)
        assert 'tfsec_config.config parameters "include" and "exclude" ' \
               'should be a list' in result[1].message
        assert isinstance(result[2], LintProblem)
        assert 'Color flag cannot be used in flags_override' \
               in result[2].message


def test_terratag():
    node_templates_content = """
    node_templates:
      cloud_resources:
        type: cloudify.nodes.terraform.Module
        properties:
          terratag_config:
            tags: { 'name_company': 'cloudify' }
            flags_override:
              - -verbose: True
              - abc: 'abc'
    """

    elem = get_mock_cfy_node(node_templates_content, 'node_templates')
    context = {
        'cloud_resources': models.NodeTemplate('cloud_resources'),
    }

    with patch('cfy_lint.yamllint_ext.rules.node_templates.ctx') as ctx:
        ctx['inputs'] = {}

        result = get_gen_as_list(
            rules.node_templates.check,
            {
                'token': elem,
                'context': context,
                'node_types': ['cloudify.nodes.terraform.Module']
            }
        )
        result.pop(0)
        assert isinstance(result[0], LintProblem)
        assert 'The flags should be without a "-" sign, -verbose' \
               in result[0].message
        assert isinstance(result[1], LintProblem)
        assert "unsupported flag, ['dir', 'skipTerratagFiles', 'verbose'," \
               " 'filter']" in result[1].message
        assert isinstance(result[2], LintProblem)
        assert "unsupported flag, {}".format(TERRATAG_SUPPORTED_FLAGS) \
               in result[2].message



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import yaml

from yamllint.parser import (
    Token,
    line_generator,
    comments_between_tokens)


class CfyToken(Token):
    def __init__(self,
                 line_no,
                 curr,
                 prev,
                 after,
                 nextnext,
                 stack,
                 blueprint_file=None):

        super().__init__(line_no, curr, prev, after, nextnext)
        self.after = self.next
        self.stack = stack
        self._node = None
        self._blueprint_file = blueprint_file

    @staticmethod
    def from_token(token):
        return CfyToken(token.line_no,
                        token.curr,
                        token.prev,
                        token.next,
                        token.nextnext)

    @property
    def node(self):
        return self._node

    @node.setter
    def node(self, value):
        self._node = value

    @property
    def blueprint_file(self):
        return self._blueprint_file

    @blueprint_file.setter
    def blueprint_file(self, value):
        self._blueprint_file = value


class CfyNode(object):

    def __init__(self, node, prev=None):
        """
        :param node: yaml.nodes.Node
        :param cfy_token: CfyToken from current or previous loop.
        """
        self.node = node
        self._prev = prev
        self._prev_prev = None
        self._next = None
        self._node_templates = None

        try:
            self._line = node.start_mark.line + 1
        except AttributeError:
            self._line = None

    @property
    def prev(self):
        return self._prev

    @prev.setter
    def prev(self, value):
        self._prev = value

    @property
    def prev_prev(self):
        return self._prev_prev

    @prev_prev.setter
    def prev_prev(self, value):
        self._prev_prev = value

    @property
    def next_token(self):
        return self._next

    @next_token.setter
    def next_token(self, value):
        self._next = value

    @property
    def node_templates(self):
        return self._node_templates

    @node_templates.setter
    def node_templates(self, value):
        self._node_templates = value

    @property
    def line(self):
        return self._line

    @line.setter
    def line(self, value):
        self._line = value


class SafeLineLoader(yaml.SafeLoader):
    def construct_mapping(self, node, deep=False):
        mapping = super(SafeLineLoader, self).construct_mapping(
            node, deep=deep)
        # Add 1 so line numbering starts at 1
        mapping['__line__'] = node.start_mark.line + 1
        return mapping


def generate_nodes_recursively(node):

    if isinstance(node, (tuple, list)):
        for sub in node:
            yield from generate_nodes_recursively(sub)
    else:
        yield node
        if isinstance(node, yaml.nodes.CollectionNode):
            for sub in node.value:
                yield from generate_nodes_recursively(sub)


def node_generator(buffer):
    yaml_loader = SafeLineLoader(buffer)
    if not yaml_loader.check_node():
        return
    yield from generate_nodes_recursively(yaml_loader.get_node().value)


def token_or_comment_generator(buffer):
    yaml_loader = SafeLineLoader(buffer)

    try:
        stack = []
        prev = None
        curr = yaml_loader.get_token()
        while curr is not None:
            next = yaml_loader.get_token()
            nextnext = (yaml_loader.peek_token()
                        if yaml_loader.check_token() else None)

            yield CfyToken(
                curr.start_mark.line + 1, curr, prev, next, nextnext, stack)

            for comment in comments_between_tokens(curr, next):
                yield comment

            prev = curr
            curr = next

    except yaml.scanner.ScannerError:
        pass


class YamlParserLintProblem():
    def __init__(self, line, column, desc):
        self.line = line
        self.column = column
        self.desc = desc


def token_or_comment_or_line_generator(buffer):
    """Generator that mixes tokens and lines, ordering them by line number"""

    tok_or_com_gen = token_or_comment_generator(buffer)
    line_gen = line_generator(buffer)
    node_gen = node_generator(buffer)

    tok_or_com = next(tok_or_com_gen, None)
    line = next(line_gen, None)
    try:
        node = CfyNode(next(node_gen, None))
    except yaml.parser.ParserError as e:
        start_line = e.context_mark.line
        end_line = e.problem_mark.line
        yield YamlParserLintProblem(
            e.context_mark.line,
            e.context_mark.column,
            '\n'.join(buffer.split('\n')[start_line:end_line])
        )
        return

    prev_node = None

    while any([g for g in [tok_or_com, line, node.node] if g is not None]):
        if should_yield_line(tok_or_com, line, node):
            yield line
            line = next(line_gen, None)
        if node.node is None or (
                tok_or_com is not None and
                node.node.start_mark.line > tok_or_com.line_no
        ):
            # while token_in_node(node, tok_or_com.line_no) is False:
            #     # We want to find a node that the token is contained in.
            #     node = next(node_gen, None)
            # if node and isinstance(tok_or_com, CfyToken):
            #     tok_or_com.node = node
            yield tok_or_com
            tok_or_com = next(tok_or_com_gen, None)
        else:
            yield node
            prev_prev_node = prev_node
            prev_node = node
            node = CfyNode(next(node_gen, None), prev_node)
            node.prev_prev = prev_prev_node


def token_in_node(node, line_no):
    try:
        return node.start_mark.line - 1 <= line_no <= node.end_mark.line - 1
    except AttributeError:
        return None


def should_yield_line(tok_or_com, line, node):
    a = tok_or_com is None or (
            line is not None and tok_or_com.line_no > line.line_no)
    b = node.node is None or (
            node is not None and node.node.start_mark.line > line.line_no)
    return a or b



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from cfy_lint.yamllint_ext.autofix.utils import filelines


def fix_empty_lines(problem):
    with filelines(problem.file) as lines:
        successive_blank_lines = 0
        index = 0
        pattern = "^ *\n"
        while re.match(pattern, lines[0]):
            lines.pop(0)

        while index < (len(lines)):
            line = lines.pop(index)
            if re.match(pattern, line):
                successive_blank_lines += 1
                if successive_blank_lines >= 2:
                    continue
            else:
                successive_blank_lines = 0
            lines.insert(index, line)
            index += 1

        while re.match(pattern, lines[-1]):
            lines.pop(-1)



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cfy_lint.logger import logger
from cfy_lint.yamllint_ext.autofix.utils import filelines

TRUE_PATTERN = 'TRUE'
FALSE_PATTERN = 'FALSE'
TRUE_REPLACEMENT = 'true'
FALSE_REPLACEMENT = 'false'


def fix_truthy(problem):
    if problem.rule == 'truthy':
        with filelines(problem.file) as lines:
            line = lines[problem.line - 1]
            line = replace_words(line, TRUE_PATTERN, TRUE_REPLACEMENT)
            line = replace_words(line, FALSE_PATTERN, FALSE_REPLACEMENT)
            lines[problem.line - 1] = line
        problem.fixed = True


def replace_words(line, pattern, replacement):
    clean_line = line.rstrip()
    new_words = []
    for word in clean_line.split(' '):
        if word.upper() == pattern:
            logger.debug('Replacing {} with {}.'.format(word, replacement))
            word = word.upper().replace(pattern, replacement)
        new_words.append(word)
    if clean_line != line:
        new_words[-1] += '\n'
    return ' '.join(new_words)



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import (setup, find_packages)

setup(
    name='cfy-lint',
    version='0.0.25',
    license='LICENSE',
    packages=find_packages(),
    description='Linter for Cloudify Blueprints',
    entry_points={
        "console_scripts": [
            "cfy-lint = cfy_lint.main:lint",
        ]
    },
    package_data={
        'cfy_lint': [
            'yamllint_ext/cloudify/__cfylint_runtime_cache/README.md',
        ]
    },
    install_requires=[
        'click',
        'pyyaml',
        'yamllint',
        'packaging'
    ]
)



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from yamllint.rules import _RULES as ruleset

from . import inputs
from . import imports
from . import node_types
from . import dsl_version
from . import capabilities
from . import relationships
from . import node_templates
from . import dsl_definitions

_CLOUDIFY_RULES = {
    inputs.ID: inputs,
    imports.ID: imports,
    node_types.ID: node_types,
    dsl_version.ID: dsl_version,
    capabilities.ID: capabilities,
    relationships.ID: relationships,
    node_templates.ID: node_templates,
    dsl_definitions.ID: dsl_definitions
}
ruleset.update(_CLOUDIFY_RULES)


def get(_id):
    if _id not in ruleset:
        raise ValueError('no such rule: "%s"' % _id)
    return ruleset[_id]
